function sol = advect(fibr,tspan)
%ADVECT Simulation of fibers in Stokes flow.
%   FIBR.SOL = ADVECT(FIBR,TSPAN) produces a 2nd order accurate
%   solution in terms of center coordinates and tangent vectors at
%   times TSPAN(1) through TSPAN(END). It is assumed that the same
%   time-step is used for the whole simulation. See FIBRSETUP for
%   details about how to setup the FIBR-structure.
%
%   The solution format is a structure containing the fields 't' and
%   'sol' where dof = sol(:,i) contains the degrees of freedom at time
%   t(i). In turn, dof(6*(j-1)+[1:6]) = [X Y Z Tx Ty Tz] (position and
%   tangent vectors, respectively) for fiber #j = 1:M.
%
%   Similarily, the velocity and angular velocity are available in the
%   field 'vel' using the format [Vx Vx Vx Wx Wy Wz].
%
%   Example:
%     fibr.Ldim = [4 4 8];
%     fibr.ntable = 5;
%     fibr.Mfibers = 10;
%     fibr = fibrsetup(fibr);
%     fibr.init = randfibr(fibr);
%     tspan = linspace(0,10,40);
%     fibr.sol = advect(fibr,tspan);
%
%     % (1) plain plot
%     figure,
%     for i = 1:numel(tspan)
%       fibrplot(fibr,tspan(i),1:fibr.Mfibers,'k');
%       grid on;
%       M(i) = getframe;
%     end
%     movie(M);
%
%     % (2) using red/blue for high/low speed
%     figure,
%     speed = compute(fibr,tspan,'v');
%     ave = mean(speed,2);
%     for i = 1:numel(tspan)
%       % fast: red, slow: blue
%       fast = speed(:,:,i) > ave(i);
%       fibrplot(fibr,tspan(i),find(fast),'r'); hold on
%       fibrplot(fibr,tspan(i),find(~fast),'b'); hold off
%       grid on;
%       M(i) = getframe;
%     end
%     movie(M);
%
%   See also FIBRSETUP, FIBRPLOT, COMPUTE, ASSEMBLE, RANDFIBR.

% S. Engblom 2010-01-28 (free space solve)
% S. Engblom 2010-01-15 (velocity output)
% S. Engblom 2009-06-08
% A-K. Tornberg (original code 'advect_fibers_per')

% quick check to avoid strange error messages
if fibr.check ~= fibrcheck(fibr)
  error('Struct has changed, call FIBRSETUP before solving.');
end

% prepare output
sol.t = tspan(:)';
sol.sol = zeros(numel(fibr.init),numel(sol.t));
sol.vel = zeros(numel(fibr.init),numel(sol.t));
difft = diff(sol.t);
dt = mean(difft);
if any(abs(difft-dt) > eps(100*(1+norm(sol.t,inf))))
  error('Can only handle constant step-size integration.');
end

% reporter, if any
if fibr.report
  report([sol.t(1) sol.t(end)], ...
         sprintf('Advecting %d fibers...',fibr.Mfibers),'init');
else
  report([],[],'none');
end

% intial data
if numel(fibr.init) ~= 6*fibr.Mfibers
  error('Incorrect number of elements in .init field.');
end
Xnew = reshape(fibr.init,6,fibr.Mfibers);
Tnew = Xnew(4:end,:);
Xnew = Xnew(1:3,:);

% check validity
Ldim = fibr.Ldim; % convenient
if ~isempty(Ldim)
  for dim = 1:3
    iout1 = find(Xnew(dim,:) < -Ldim(dim)/2);
    Xnew(dim,iout1) = Xnew(dim,iout1)+Ldim(dim);
    iout2 = find(Xnew(dim,:) > Ldim(dim)/2);
    Xnew(dim,iout2) = Xnew(dim,iout2)-Ldim(dim);
    if ~isempty(iout1) || ~isempty(iout2)
    warning('Wrapping: coordinates of .init field outside periodic box.');
    end
  end
end
nrm = realsqrt(sum(Tnew.*Tnew));
if any(abs(nrm-1) > 1e-10)
  warning('Renormalizing tangents in .init field.')
  Tnew = tprod(1./nrm,Tnew,[3 2],[1 2]);
end

% insert initial data
i = 1;
sol.sol(:,i) = reshape([Xnew; Tnew],[],1);
i = i+1;

% advect!
Vnew = [];
Rnew = [];
for tt = sol.t(1:end-1)
  Vold = Vnew;
  Rold = Rnew;
  [Vnew,Rnew] = velocity(fibr,Xnew,Tnew);
  if i > 2
    % 2nd order scheme
    Xtmp = 4/3*Xnew-1/3*Xold+2/3*dt*(2*Vnew-Vold);
    Ttmp = 4/3*Tnew-1/3*Told+2/3*dt*(2*Rnew-Rold);
 
    if ~isempty(Ldim)
      % periodic boundaries when the two latest steps are outside
      for dim = 1:3
        iout = find(Xtmp(dim,:) < -Ldim(dim)/2 & Xnew(dim,:) < -Ldim(dim)/2);
        Xtmp(dim,iout) = Xtmp(dim,iout)+Ldim(dim);
        Xnew(dim,iout) = Xnew(dim,iout)+Ldim(dim);
        iout = find(Xtmp(dim,:) > Ldim(dim)/2 & Xnew(dim,:) > Ldim(dim)/2);
        Xtmp(dim,iout) = Xtmp(dim,iout)-Ldim(dim);
        Xnew(dim,iout) = Xnew(dim,iout)-Ldim(dim);
      end
    end
  else
    % 1st step, 1st order Euler forward
    Xtmp = Xnew+dt*Vnew;
    Ttmp = Tnew+dt*Rnew;
  end
  Xold = Xnew;
  Xnew = Xtmp;
  Told = Tnew;
  Tnew = Ttmp;

  % normalize
  % *** questionable!
  % fprintf('Norm %f\n',norm(realsqrt(sum(Tnew.*Tnew))-1,inf));
  Tnew = tprod(Tnew,1./realsqrt(sum(Tnew.*Tnew)),1:2,[3 2]);

  % save data and report progress
  sol.sol(:,i) = reshape([Xnew; Tnew],[],1);
  sol.vel(:,i-1) = reshape([Vnew; Rnew],[],1); % NB: previous step i-1
  i = i+1;
  report(tt,[],'');
end

% nice to obtain a "full" solution:
[Vnew,Rnew] = velocity(fibr,Xnew,Tnew);
sol.vel(:,end) = reshape([Vnew; Rnew],[],1);
 
% close the reporter
report([],[],'done');
