function [A,B,V,R,CV,CR] = assemble(xc,tv,Ldim,eeps,F,G,p,wL,M2L,aux)
%ASSEMBLE Integral equations for fibers in Stokes flow.
%   [A,B,V,R,CV,CR] = ASSEMBLE(XC,TV,Ldim,eeps,F,G,P,wL,M2L,aux)
%   assembles the integral equations for the force expansion X in AX =
%   B for a collection of fibers with center coordinates XC and
%   tangent vectors TV, both arguments with dimensions 3-by-M, where M
%   is the number of fibers. Additionally, the tangential- and
%   rotational velocities V and R are (partially) assembled so that
%   the total velocities are given by Vtotal = V+CV*X and Rtotal =
%   R+CR*X, where X = A\B is the force expansion.
%
%   The geometry is either a periodic box Ldim = [Lx Ly Lz] or the
%   problem is in free space with Ldim empty. Stokes flow with radius
%   of each fiber given by 2*eeps and external forces F is
%   understood. For the periodic case, the Green's function remainder
%   is tabulated in G (see GREENTABLE). P and wL is the quadrature and
%   values of Legendre polynomials, respectively (see LEGQUAD and
%   LEGEVAL) and M2L is a monomial to Legendre transformation as
%   computed by FIBRSETUP. The last argument aux contains some
%   auxiliary constants, see FIBRSETUP.
%
%   Performance of this function is central. No error-checking is
%   performed.
%
%   See also FIBRSETUP, VELOCITY.

% S. Engblom 2010-01-28 (free space assembly added)
% S. Engblom 2009-06-05 (Revision)
% A-K. Tornberg (original codes 'assemble_matrix_per',
% 'assemble_rhs_per' and 'compute_velocities_per')

% dimensions
M = size(xc,2);
N = size(wL,2)-1;

% quadrature on each fiber
xi = tsum(p*tv(1,:),xc(1,:),[1 2],[3 2]);
yi = tsum(p*tv(2,:),xc(2,:),[1 2],[3 2]);
zi = tsum(p*tv(3,:),xc(3,:),[1 2],[3 2]);

% allocation
A = zeros(3*M*N);
B = zeros(3*M*N,1);
V = zeros(3,M);
R = zeros(3,M);
CV = zeros(3*M,3*M*N);
CR = zeros(3*M,3*M*N);

% used to save previous interactions and only compute unique
% interactions (i,j) with i >= j
Theta = zeros((N+1)*M,(N+1)*M,6);

for j = 1:M
  % contribution from single fibre j...
  xj = xc(:,j);
  tj = tv(:,j);
  Fj = F(:,j);
  V(:,j) = V(:,j)+0.5/aux.d*((aux.d+2)*Fj+(aux.d-2)*tj'*Fj*tj);

  % ...to collection of fibers i

  if ~isempty(Ldim)
    % quadrature points xxi, yyi, and zzi such that the closest periodic
    % image of the fiber is used (as measured at the midpoint)
    rd = tsum(xc(1,j:end)-xj(1),[-Ldim(1); 0; Ldim(1)],[3 2],1);
    [foo,imin] = min(abs(rd),[],1);
    xxi = tsum(xi(:,j:end),(imin-2)*Ldim(1),[1 2],[3 2]);
    rd = tsum(xc(2,j:end)-xj(2),[-Ldim(2); 0; Ldim(2)],[3 2],1);
    [foo,imin] = min(abs(rd),[],1); 
    yyi = tsum(yi(:,j:end),(imin-2)*Ldim(2),[1 2],[3 2]);
    rd = tsum(xc(3,j:end)-xj(3),[-Ldim(3); 0; Ldim(3)],[3 2],1);
    [foo,imin] = min(abs(rd),[],1); 
    zzi = tsum(zi(:,j:end),(imin-2)*Ldim(3),[1 2],[3 2]);
    
    % scalar product of Green's function with basis polynomials
    iG = reshape(greendotleg2(xxi,yyi,zzi,xj,tj,G,p,wL),[],M-j+1,N+1,6);
    % remove the case i = j to avoid the singularity:
    iG(:,2:end,:,:) = iG(:,2:end,:,:)+ ...
        reshape(greendotleg1(xxi(:,2:end),yyi(:,2:end),zzi(:,2:end), ...
                             xj,tj,eeps,N,M2L),size(iG)-[0 1 0 0]);
  else
    % free space
    xxi = xi(:,j:end);
    yyi = yi(:,j:end);
    zzi = zi(:,j:end);

    iG = reshape(greendotleg1(xxi(:,2:end),yyi(:,2:end),zzi(:,2:end), ...
                              xj,tj,eeps,N,M2L),size(xxi,1),M-j,N+1,6);
    iG = cat(2,zeros(size(iG,1),1,N+1,6),iG);
  end

  % integrate with basis polynomials
  iG = reshape(wL'*reshape(iG,size(wL,1),[]),N+1,M-j+1,N+1,6);

  % save iG to Theta, symmetrize it and reload to iG in order to compute
  % only unique interactions
  col = (N+1)*(j-1)+(1:N+1);
  Theta((N+1)*(j-1)+1:end,col,:) = reshape(iG,[],size(col,2),6);
  Theta(col,(N+1)*j+1:end,:) = permute(Theta((N+1)*j+1:end,col,:), ...
                                       [2 1 3]);
  iG = reshape(Theta(:,col,:),N+1,M,N+1,6);

  % expand iG to "full symmetric" format
  iG = reshape(iG(:,:,:,[1 4 6 4 2 5 6 5 3]),N+1,M,N+1,3,3);

  % general assembly
  thetaG = tprod(iG(2:end,:,:,:,:),aux.gamma,[2 3 5 1 4],2);
  ip = tprod(thetaG,tv,[-1 1:4],[-1 2]); % inner product (thetaG,tv)
  ip = tprod(tprod(aux.ee,ip,1,1:4),tv,2:5,[1 3]);
  thetaG(:,1,:,:,:) = 0; % special case
  thetaG = reshape(thetaG-ip,[],3*(N+1));

  % (1) columns of A: order 1..N
  col = 3*N*(j-1)+(1:3*N);
  A(:,col) = thetaG(:,4:end);

  % (2) contribution to B: 0th order polynomial only
  B = B-thetaG(:,1:3)*0.5*Fj(:);

  % (3) contributions to V and R and blocks in CV and CR: order 0..N
  theta0 = tprod(iG(1:2,:,1,:,:),0.5*Fj(:),[3 2 4 1 -1],-1);
  theta1 = theta0(:,:,2);
  theta0 = theta0(:,:,1);
  V = V+0.5/aux.d*theta0;
  R = R+1.5/aux.d*(theta1-tprod(sum(tv.*theta1),tv,[3 2],[1 2]));

  % columns of CV and CR
  theta0 = permute(iG(1:2,:,2:end,:,:),[4 2 5 3 1]);
  theta1 = theta0(:,:,:,:,2);
  theta0 = theta0(:,:,:,:,1);
  CV(:,col) = CV(:,col)+reshape(0.5/aux.d*theta0,[],3*N);
  CR(:,col) = CR(:,col)+ ...
      reshape(1.5/aux.d*(theta1- ...
                         tprod(tprod(tv,theta1,[-1 1],[-1 1 2 3]), ...
                               tv,[2 3 4],[1 2])),[],3*N);
end

% final (diagonal) contribution
A = A+eye(size(A));
