function varargout = compute(fibr,t,varargin)
%COMPUTE Compute physical quantities of fibers.
%   [out1,out2,...] = COMPUTE(FIBR,T,'prop 1','prop 2', ...)  computes
%   the values out1, out2, ... at times T of the properties 'prop 1',
%   'prop 2', ... The output arrays have dimensions
%   dim-by-FIBR.Mfibers-by-numel(T), where the values of dim are
%   listed in the table of supported properties below.
%
%   [...] = COMPUTE(FIBR,T,...,IX) computes properties for fibers with
%   index IX only.
%
%   Property      dim    Description
%   -----------------------------------------------------------------------
%   'r'           3      Coordinate of the midpoint.
%   't'           3      Tangential vector.
%   'dr'          3      Velocity of the midpoint.
%   'dt'          3      Tangential velocity.
%   'v'           1      Midpoint speed (norm of the velocity).
%
%   'P'           3      Linear momentum about origo.
%   'l'           3      Angular momentum about each fiber's center
%                        of mass
%   'L'           3      Angular momentum about origo.
%
%   'Ekin'        1      Kinetic energy.
%   'Erot'        1      Rotational energy.
%   'Epot'        1      Potential energy.
%   'E'           1      Total energy.
%
%   Example:
%     % position and speed at time t = 0
%     [r,v] = compute(fibr,0,'r','v');
%
%     % tangents for first 5 fibers
%     t = compute(fibr,0,'t',1:5); 
%
%   See also ADVECT.

% S. Engblom 2010-01-26

% parse
if ~isnumeric(varargin{end})
  % index to fibers not given
  out = varargin;
  ix0 = 1:fibr.Mfibers;
else
  % index given (last in varargin)
  out = varargin(1:end-1);
  ix0 = varargin{end};
end
ix = reshape(tsum(6*(ix0(:)-1),1:6,2,[3 1]),[],1);

% interpolate
sol = interp1(fibr.sol.t,fibr.sol.sol(ix,:)',t,'linear')';
sol = reshape(sol,6,size(ix,1)/6,[]);
R = sol(1:3,:,:);
T = sol(4:6,:,:);
vel = interp1(fibr.sol.t,fibr.sol.vel(ix,:)',t,'linear')';
vel = reshape(vel,6,size(ix,1)/6,[]);
dR = vel(1:3,:,:);
dT = vel(4:6,:,:);

% parse properties and produce output
varargout = cell(1,numel(out));
for i = 1:numel(out)
  switch out{i}
   case 'r'
    varargout{i} = R;
   case 't'
    varargout{i} = T;
   case 'dr'
    varargout{i} = dR;
   case 'dt'
    varargout{i} = dT;
   case 'v'
    varargout{i} = sqrt(sum(dR.*dR,1));
   case 'P'
    varargout{i} = dR;
   case 'l'
    varargout{i} = 2/3*cross(T,dT,1);
   case 'L'
    varargout{i} = dR+2/3*cross(T,dT,1);
   case 'Ekin'
    varargout{i} = 1/2*sum(dR.*dR,1);
   case 'Erot'
    varargout{i} = 1/6*sum(dT.*dT,1);
   case 'Epot'
    varargout{i} = permute( ...
        -tprod(fibr.ExtForce(:,ix0),R, ...
               [-1 1],[-1 1 2]),[3 1 2]);
   case 'E'
    varargout{i} = 1/2*sum(dR.*dR,1)+ ...
        1/6*sum(dT.*dT,1)+ ...
        permute( ...
            -tprod(fibr.ExtForce(:,ix0),R, ...
                   [-1 1],[-1 1 2]),[3 1 2]);
   otherwise
    error(['Unknown property ''' out{i} '''.']);
  end
end
