function D = distances(xtdof1,xtdof2,rlim)
%DISTANCES Intra-fiber distances.
%   D = DISTANCES(XTdof1,XTdof2,rlim) computes all distances less than
%   rlim for fibers with coordinates given by the 6-by-M1 matrix
%   XTdof1 and the 6-by-M2 matrix XTdof2, respectively. In both cases
%   XTdof(1:3,j) are the coordinates of the midpoint of fiber j and,
%   similarily, XTdof(4:6,j) contains tangent vectors.
%
%   On return, D is a sparse matrix of size M1-by-M2 with D(i,j) =
%   distance between fiber XTdof1(:,i) and XTdof2(:,j), or 0 if the
%   distance is >= rlim.
%
%   D = DISTANCES(XTdof,rlim) assumes XTdof1 = XTdof2 and runs
%   faster. In this case only the strictly lower triangular part of D
%   is returned.
%
%   No error-checking is performed.

% S. Engblom 2010-02-01
% A-K. Tornberg 2009-02-25 (original codes 'rod_dist' and
% 'filament_distances')

% The distance check is made in three steps as follows;
%
% (1) Check the distance between the center coordinates. If this
% distance is at least 2+rlim, the distance need not be computed.
%
% (2) Put Nsub+1 discrete points on each remaining fiberpair and check
% the minimum distance between all such points. Trust this result if
% the computed distance is larger than sqrt(2)*Lsub/2 = Lsub/sqrt(2),
% where Lsub = 2/Nsub.
%
% (3) Otherwise and finally, use l_rdist() to compute each remaining
% distance exactly.

if nargin == 3
  Dijs = zeros(0,3); % index and values of sparse matrix D
  x1 = xtdof1(1:3,:);
  x2 = xtdof2(1:3,:);

  % (1) coarsest check:
  rv = realsqrt(sum(tsum(x1,-x2,[3 1],[3 2]).^2,3));
  [i1,j1] = find(rv <= 2+rlim);
  i1 = reshape(i1,1,[]); % stupid bug in find!
  j1 = reshape(j1,1,[]);
  [iu,foo,ib] = fsetop('unique',i1);
  [ju,foo,jb] = fsetop('unique',j1);

  if ~isempty(i1)
    % (2) more careful check: create vectors with Nsub+1 points on each
    % critical pair (i1,j1) of fibers
    Nsub = 8; % *** prefer 7 or 9? (sv = 0 has already been used...)
    Lsub = 2/Nsub;
    dlim = max(Lsub/sqrt(2),rlim);
    sv = linspace(-1,1,Nsub+1);
    t1 = xtdof1(4:6,:);
    t2 = xtdof2(4:6,:);
    X1 = tsum(x1(:,iu),tprod(t1(:,iu),sv,[1 2],[4 3]), ...
              [1 2],[1 2 3]);
    X2 = tsum(x2(:,ju),tprod(t2(:,ju),sv,[1 2],[4 3]), ...
              [1 2],[1 2 3]);

    % compute the minimum distance from all those points
    rv = realsqrt(sum(tsum(X1(:,ib,:),-X2(:,jb,:),[1 2 3],[1 2 4]).^2));
    rv = min(min(rv,[],4),[],3);
    ij = find(rv <= dlim);
    i2 = i1(ij);
    j2 = j1(ij);

    if ~isempty(i2)
      % (3) the point-based tests could not conlude, do a full calculation
      l = 0;
      for k = 1:numel(i2)
        d = l_rdist(x1(:,i2(k)),t1(:,i2(k)),[-1 1], ...
                    x2(:,j2(k)),t2(:,j2(k)),[-1 1]);
        if d < rlim
          l = l+1;
          Dijs(l,:) = [i2(k) j2(k) d];
        end
      end
    end
  end
  D = sparse(Dijs(:,1),Dijs(:,2),Dijs(:,3),size(x1,2),size(x2,2));
else
  rlim = xtdof2;
  Dijs = zeros(0,3);
  x1 = xtdof1(1:3,:);
  M = size(xtdof1,2);

  % same idea, but twice as fast
  [i1,j1] = find(tril(ones(M),-1)); % unique pairs only
  rv = realsqrt(sum((x1(:,i1)-x1(:,j1)).^2));
  ij = find(rv <= 2+rlim);
  i1 = reshape(i1(ij),1,[]);
  j1 = reshape(j1(ij),1,[]);
  iju = fsetop('union',i1,j1);
  [foo,ib] = fsetop('ismember',i1,iju);
  [foo,jb] = fsetop('ismember',j1,iju);

  if ~isempty(i1)
    Nsub = 8; % *** 7 or 9?
    Lsub = 2/Nsub;
    dlim = max(Lsub/sqrt(2),rlim);
    sv = linspace(-1,1,Nsub+1);
    t1 = xtdof1(4:6,:);
    X1 = tsum(x1(:,iju),tprod(t1(:,iju),sv,[1 2],[4 3]), ...
              [1 2],[1 2 3]);

    rv = realsqrt(sum(tsum(X1(:,ib,:),-X1(:,jb,:),[1 2 3],[1 2 4]).^2));
    rv = min(min(rv,[],4),[],3);
    ij = find(rv <= dlim);
    i2 = i1(ij);
    j2 = j1(ij);

    if ~isempty(i2)
      l = 0;
      for k = 1:numel(i2)
        d = l_rdist(x1(:,i2(k)),t1(:,i2(k)),[-1 1], ...
                    x1(:,j2(k)),t1(:,j2(k)),[-1 1]);
        if d < rlim
          l = l+1;
          Dijs(l,:) = [i2(k) j2(k) d];
        end
      end
    end
  end
  D = sparse(Dijs(:,1),Dijs(:,2),Dijs(:,3),M,M);
end

%-------------------------------------------------------------------------
function rdist = l_rdist(x1,t1,srange1,x2,t2,srange2)
%L_DIST Distance between two straight rods.
%  RDIST = L_RDIST(X1,T1,SRANGE1,X2,T2,SRANGE2) computes the Euclidean
%  distance between the two rods X1+s*T1, s in (SRANGE(1),SRANGE(2))
%  and X2+s*T2, s in (SRANGE(1),SRANGE(2)).
%
%  It is assumed that all vectors are column vectors except for
%  SRANGE1 and SRANGE2 which should be rows. Also, the tangent vectors
%  must be normalized.

% ensure inner product is nonnegative
aa = t1'*t2;
if aa < 0 
  t2 = -t2;
  srange2 = -srange2([2 1]);
  aa = -aa;
end

% endpoints
p1 = tsum(x1,t1*srange1,1,[1 2]);
p2 = tsum(x2,t2*srange2,1,[1 2]);

% various distances
Rvec0 = x2-x1;
Rvec = p2(:,[1 2 1 2])-p1(:,[1 1 2 2]);
Rmod = sum(Rvec.^2);
Rmode = min(Rmod);

% parallel rods (but not anti-parallel thanks to the initial flip)
if abs(aa-1) < eps(1000)
  % normal distance
  rdist = norm(Rvec0-(Rvec0'*t1)*t1);
  if rdist < eps(1000) % (some cancellation in x2-x1 may occur)
    % along the same line
    rdist = 0;
  end

  % end-to-end vectors projected onto t1
  dd = Rvec'*t1;

  % in this case the rods are not overlapping in the sense that no vector
  % in the normal direction from one rod can hit the other
  if dd(1)*dd(3) > 0 && dd(2)*dd(4) > 0
    rdist = realsqrt(Rmode);
  end
else
  % not parallel, use formula
  Rmod0 = norm(x2-x1)^2;
  b1 = Rvec0'*t1;
  b2 = Rvec0'*t2;

  % tentative values of parameter s
  s1 = (b1-aa*b2)/(1-aa.^2);
  s2 = (aa*b1-b2)/(1-aa.^2);
  
  % distance formula
  if srange1(1) < s1 && s1 < srange1(2) && ...
        srange2(1) < s2 && s2 < srange2(2)
    % both are inside their intervals
    Rmod2 = Rmod0+s1.^2+s2.^2-2*s1*s2*aa-2*s1*b1+2*s2*b2;
  else
    % try optimal value of s2 conditioned on s1 within limits
    ss1 = median([srange1 s1]);
    ss2 = aa*ss1-b2;
    if srange2(1) < ss2 && ss2 < srange2(2) 
      Rmod2a = Rmod0+ss1.^2+ss2.^2-2*ss1*ss2*aa-2*ss1*b1+2*ss2*b2;
    else
      Rmod2a = Rmode;
    end

    % and vice versa
    ss2 = median([srange2 s2]);
    ss1 = aa*ss2+b1;
    if srange1(1) < ss1 && ss1 < srange1(2) 
      Rmod2b = Rmod0+ss1.^2+ss2.^2-2*ss1*ss2*aa-2*ss1*b1+2*ss2*b2;
    else
      Rmod2b = Rmode;
    end

    % both might work, take the smallest
    Rmod2 = min(Rmod2a,Rmod2b);
  end
  
  rdist = realsqrt(Rmod2);
end

%-------------------------------------------------------------------------
