function fibr = fibrsetup(varargin)
%FIBRSETUP Setup simulation of fibers in Stokes flow.
%   FIBR = FIBRSETUP(FIBR) extends the structure FIBR with various
%   fields so that the resulting structure can be used with ADVECT.
%
%   FIBR = FIBRSETUP('FIELD1',VAL1,...) is a convenient way of
%   achieving the same thing.
%
%   The most important fields are summarized below. To see a list of
%   all fields, type 'more on, type fibrsetup' at the Matlab
%   prompt. To get all the default values, simply type 'fibrsetup'.
%
%   FIBRSETUP tries to reuse fields that are expensive to compute. In
%   particular, the Green's function table is reused provided that the
%   format is correct (field 'Green').
%
%   Property  Value/{Default}     Description
%   -----------------------------------------------------------------------
%   Ldim      3-vector {[4 4 4]}  Geometry of the periodic domain,
%                                 use empty for free space.
%
%   eeps      Scalar {0.01}       Slender body approximation
%                                 parameter (2*radius/fiber length).
%
%   Mfibers   Scalar {10}         Number of fibers.
%
%   ExtForce  3-by-Mfibers matrix External forces. The default is a
%                                 downward unit gravity force.
%
%   report    Boolean {true}      Report while solving.
%
%   Nforce    Scalar {5}          Number of terms in the Legendre
%                                 force expansion.
%
%   nquad     Length 2 vector     Order and number of quadratures used 
%             {[3 8]}             (see LEGQUAD).
%
%   ntable    Scalar or length 2  Order and stepsize in Green's function
%             vector {[2 0.2]}    table: ntable(2) is ignored
%                                 unless ntable(1) is 2 (piecewise
%                                 linear) or 4 (piecewise cubic). A
%                                 single scalar means that
%                                 Chebyshev polynomials up to order
%                                 ntable-1 are used.
%
%   Example:
%     % "minimal" simulation:
%     fibr = fibrsetup(struct); % all defaults used
%     fibr.init = randfibr(fibr);
%     fibr.sol = advect(fibr,0:0.1:1);
%     for t = fibr.sol.t, fibrplot(fibr,t); pause(0.1); end
%
%   See also ADVECT, ASSEMBLE.

%   The complete format of the FIBR-structure:
%
%   Field     Value/{Default}     Description
%   -----------------------------------------------------------------------
%   Ldim      3-vector {[4 4 4]}  Geometry of the domain, empty for
%                                 free space.
%
%   eeps      Scalar {0.01}       Slender body approximation
%                                 parameter (2*radius/fiber length)
%
%   Mfibers   Scalar {10}         Number of fibers.
%
%   ExtForce  3-by-Mfibers matrix External forces. The default is a
%                                 unit gravity force.
%
%   report    Boolean {true}      Report while solving.
%
%   Nforce    Scalar {5}          Number of terms in the Legendre
%                                 force expansion.
%
%   nquad     Length 2 vector     Order and number of quadratures used 
%             {[3 8]}             (see LEGQUAD).
%
%   ntable    Scalar or length 2  Order and stepsize in Green's function
%             vector {[2 0.2]}    table.
%
%   rlim      Scalar {0.1}        Minimal initial distance between
%                                 fibers. See RANDFIBR.
%
%   Green     Scalar struct       Green's function remainder table,
%   .G                            see GREENTABLE and GREENSTOKES.
%   .sizG
%   .Lmax
%   .ntable
%   .check
%
%   Legendre  Scalar struct       Legendre polynomials.
%   .w
%   .p
%   .L        (currently unused)
%   .wL
%   .M2L
%
%   aux       Scalar struct       Auxiliary variables.
%   .d
%   .ee
%   .gamma
%
%   init      6-by-Mfibers Matrix Initial data, see RANDFIBR.
%
%   sol       Scalar struct       Solution structure, see ADVECT.
%   .t
%   .sol
%   .vel
%
%   check     Scalar uint32       Checksum, see FIBRCHECK.

% S. Engblom 2010-01-28 (free space syntax added)
% S. Engblom 2009-06-08
% A-K. Tornberg (original code 'advect_fibers_per')

% defaults
fibrdef.Ldim = [4 4 4];
fibrdef.eeps = 0.01;
fibrdef.Mfibers = 10;
fibrdef.ExtForce = []; % set later

fibrdef.report = true;

fibrdef.ntable = [2 0.2];
fibrdef.Nforce = 5;
fibrdef.nquad = [3 8];
fibrdef.rlim = 0.1;

% add various empty fields since it is nice to be able to use an old
% struct without getting errors
fibrdef.Green = struct('G',[],'sizG',[],'Lmax',[], ...
                       'check',[],'ntable',[]);

% polynomials
fibrdef.Legendre = struct('p',[],'w',[],'L',[],'wL',[],'M2L',[]);

% auxiliaries
fibrdef.aux = struct('d',[],'ee',[],'gamma',[]);

% solution fields
fibrdef.init = [];
fibrdef.sol = struct('sol',[],'vel',[],'t',[]);

% checksum
fibrdef.check = []; % set later

% special call: return defaults
if nargin == 0, fibr = fibrdef; return; end

% merge defaults with actual inputs
if nargin == 1
  [fibr,got] = parseopts(fibrdef,varargin{1});
else
  [fibr,got] = parseopts(fibrdef,varargin);
end

% convenient
M = fibr.Mfibers;
N = fibr.Nforce;

% check given fields and generate missing ones
if got.ExtForce && ~isequal(size(fibr.ExtForce),[3 M])
  warning('Incorrect format of external forces: using default.');
  got.ExtForce = 0;
end
if ~got.ExtForce
  fibr.ExtForce = zeros(3,M);
  fibr.ExtForce(3,:) = -1; % unit gravity force
end

% quadrature
if got.nquad && isscalar(fibr.nquad)
  fibr.nquad = [fibr.nquad 1];
end
[fibr.Legendre.p,fibr.Legendre.w] = legquad(fibr.nquad(1),fibr.nquad(2));
fibr.Legendre.L = legeval(fibr.Legendre.p,N)';
fibr.Legendre.wL = tprod(fibr.Legendre.w,fibr.Legendre.L,1,[1 2]);

% monomial to Legendre lower triangular transformation matrix;
% M2L(i,j) are the monomial basis coefficients of the ith Legendre
% polynomial
M2L = zeros(N+1);
M2L(1,1) = 1;
if N > 0
  M2L(2,2) = 1;
  for i = 3:N+1
    M2L(i,1) = -(i-2)/(i-1)*M2L(i-2,1); 
    for j = 2:i
      M2L(i,j) = (2*i-3)/(i-1)*M2L(i-1,j-1)-(i-2)/(i-1)*M2L(i-2,j);
    end
  end
end
fibr.Legendre.M2L = M2L;

% periodic BCs if non-empty
if ~isempty(fibr.Ldim)
  % Green's table (try to reuse if possible)
  if fibr.ntable(1) ~= 2 && fibr.ntable(1) ~= 4
    fibr.ntable = fibr.ntable(1);
  end

  % old table? is it ok?
  if got.Green.G && got.Green.check && ...
        (isempty(fibr.Green.check) || ...
         fibr.Green.check ~= fsetop('check',[fibr.Ldim(:); fibr.Ldim(:)/2+2; ...
                        fibr.ntable(:)]))
    warning('Cannot use input Green''s function.');
    got.Green.G = 0;
  elseif got.Green.G && ~got.Green.check
    % table explicitly set; mark it with a checksum for next time
    fibr.Green.check = fsetop('check',[fibr.Ldim(:); fibr.Ldim(:)/2+2; ...
                        fibr.ntable(:)]);

    % used to interpret the format:
    fibr.Green.ntable = fibr.ntable;
  end

  if got.Green.G
    % got the table, will try to use it
    if fibr.report
      fprintf('Using existing Green''s function.\n');
    end
  elseif ~isscalar(fibr.ntable)
    % constant step table
    if fibr.report
      fprintf('Tabulating Green''s function.\n');
      if fibr.ntable(1) == 2
        fprintf('  Linear interpolation.\n');
      elseif fibr.ntable(1) == 4
        fprintf('  Cubic interpolation.\n');
      else
        error('Unknown interpolation.');
      end
    end
    % tabulating one fiber length further than Ldim/2 for the extreme case
    % where the centers are separated by Ldim/2
    fibr.Green = greentable(fibr.Ldim,fibr.Ldim/2+2, ...
                            fibr.ntable,'report',fibr.report);
  else
    % Chebyshev coefficients
    if fibr.report
      fprintf('Computing Chebyshev approximation of Green''s function.\n');
    end
    fibr.Green = greentable(fibr.Ldim,fibr.Ldim/2+2, ...
                            fibr.ntable,'report',fibr.report);
  end
end

% auxiliaries used by assemble
fibr.aux.d = -1-2*reallog(fibr.eeps);
lambda = cumsum(2./(1:N))';
fibr.aux.ee = [-0.75/(fibr.aux.d-2); ...
               (fibr.aux.d-2-lambda(2:end))/2./ ...
               (fibr.aux.d-lambda(2:end))];
fibr.aux.gamma = [1; ((2:N)'+0.5)./ ...
                  (fibr.aux.d+2-lambda(2:end))];

% display some information
if fibr.report
  fprintf('M = %d fibers and N = %d terms in the force expansion.\n', ...
          M,N);   
end

% compute checksum
fibr.check = fibrcheck(fibr);
