function iG = greendotleg1(x,y,z,xc,tv,eeps,N,M2L)
%GREENDOTLEG1 Green's function times Legendre basis (scalar product).
%   IG = GREENDOTLEG1(X,Y,Z,XC,TV,eeps,N,M2L) computes the integral
%   inner product
%
%     IG = int G([X Y Z]-s) L(s) ds,
%
%   where the integral is over a fiber with center coordinates XC,
%   tangent vector TV and slenderness parameter eeps. N is the highest
%   order of the Legendre polynomials and M2L is a monomial to
%   Legendre transformation matrix. G is the center Stokeslet: for
%   numerically computing this integral with respect to the periodic
%   remainder, see GREENDOTLEG2.
%
%   The output IG is a numel(X)-by-(N+1)-by-6 3-D array.
%
%   Performance of this function is important. Only limited
%   error-checking is performed.
%
%   See also GREENDOTLEG2, GREENINT.

% S. Engblom 2009-05-11 (Revision)
% A-K. Tornberg (original code 'int_analytical_vec')

% normalize input
x = x(:)-xc(1);
y = y(:)-xc(2);
z = z(:)-xc(3);
tv = tv(:);

X = [x y z];
X2 = [x.*x y.*y z.*z x.*y y.*z x.*z];

% analytically computed integrals (radius a = 2*eeps, the 'max' below
% ensures that the distance cannot be smaller than this)
iP = greenint(-2*X*tv,max(sum(X2(:,1:3),2),4*eeps^2),M2L);

% assembly...
eps2 = 2*eeps^2;
iP1 = iP(:,:,1)+eps2*iP(:,:,2);
iP2 = iP(:,:,2)-3*eps2*iP(:,:,3);
iP4 = iP(:,:,4)-3*eps2*iP(:,:,5);
iP6 = iP(:,:,6)-3*eps2*iP(:,:,7);
tv1 = [tv; tv];
tv2 = [tv; tv([2 3 1])];
Xtv = tprod(X(:,[1:3 1:3]),tv2,[1 2],2)+ ...
      tprod(X(:,[1:3 2 3 1]),tv1,[1 2],2);

% ...final formula
iG = tprod(iP2,X2,[1 2],[1 3])- ...
     tprod(iP4,Xtv,[1 2],[1 3])+ ...
     tprod(iP6,tv1.*tv2,[1 2],3);
iG(:,:,1:3) = tsum(iG(:,:,1:3),iP1,[1 2 3],[1 2]);
