function [iG,jG] = greenint(b,c,M2L)
%GREENINT Analytic integral of Green's function times Legendre basis.
%   IG = GREENINT(b,c,M2L) computes the set of integrals
%
%     IG = int s^alpha L(s)/sqrt(s^2+bs+c)^beta ds,
%
%   for alpha = [0 1 2] and beta = [1 3 5], where L(s) are the
%   Legendre basis polynomials up to and including order N. M2L is a
%   (N+1)-by-(N+1) precomputed monomial to Legendre transformation
%   matrix.
%
%   On return, IG is numel(c)-by-(N+1)-by-7 and the third dimension
%   enumerates the cases (alpha,beta) = [(0,1),(0,3),(0,5),(1,3),
%   (1,5),(2,3),(2,5)] of the integral.
%
%   See also GREENDOTLEG1.

%   [IG,JG] = GREENINT(b,c,M2L) computes the same thing twice, but
%   uses a different recurrence. For experimental purposes.

% S. Engblom 2009-11-24 (Revision)
% A-K. Tornberg (original code 'Pk_intergals_vec')

% polynomial orders 0:N
N = size(M2L,1)-1;

% discriminant
d = c-0.25*b.*b;
if any(d < 0)
  error('Negative discriminant not allowed.');
end
ub = realsqrt(1+b+c);
ua = realsqrt(1-b+c);

% homogeneous solution grows as sqrt(c)^n, use forward- or backward
% recursion to minimize errors
P = 2;
cmax = (2/eps/(P*(N+3)+2))^(2/(P*(N+3)+1));
% (formula derived by assuming that the initial data has error ~ eps,
% that the backward recurrence starts from k = P*(N+3) and equates the
% worst forward/backward errors under the estimate I1(k) ~
% 2/(2k+1)/sqrt(c), to solve for the cutoff value of c)
ifwd = find(c <= cmax);
ibwd = find(cmax < c);
b1 = b(ifwd);
b2 = b(ibwd);
c1 = c(ifwd);
c2 = c(ibwd);
ua1 = ua(ifwd);
ua2 = ua(ibwd);
ub1 = ub(ifwd);
ub2 = ub(ibwd);

% allocate
len = numel(c);
I1 = zeros(len,N+2);
I3 = zeros(len,N+3);
I5 = zeros(len,N+3);

% initial two first values
I1(:,1) = reallog(abs((2+b+2*ub)./(-2+b+2*ua)));
I1(:,2) = (ub-ua)-b/2.*I1(:,1);

% initial data for I3 and I5 changes when the discriminant is zero
izero = find(abs(d) < 1e-14); 
d(izero) = 1e-14;

I3(:,1) = (1+0.5*b)./(d.*ub)-(-1+0.5*b)./(d.*ua);
I3(izero,1) = -0.5./(1+0.5*b(izero)).^2+ ...
    0.5./(-1+0.5*b(izero)).^2;
I3(:,2) = -1./ub+1./ua-0.5*b.*I3(:,1);

I5(:,1) = (1+0.5*b)./(3*d.*ub.^3)-(-1+0.5*b)./(3*d.*ua.^3)+ ...
          2./(3*d).*I3(:,1);
I5(izero,1) = -0.25./(1+0.5*b(izero)).^4+0.25./(-1+0.5*b(izero)).^4;
I5(:,2) = -(0.5*b+c)./(3*d.*ub.^3)+(-0.5*b+c)./(3*d.*ua.^3)- ...
          b./(3*d).*I3(:,1);
I5(izero,2) = -1./(3*(1+0.5*b(izero)).^3)+1./(3*(-1+0.5*b(izero)).^3)- ...
    0.5*b(izero).*I5(izero);

if ~isempty(ifwd)
  fac = -1;
  % forward recursion...
  for n = 2:N+1
    I1(ifwd,n+1) = (ub1-fac*ua1)/n+ ...
        (1-2*n)/(2*n)*b1.*I1(ifwd,n)-(n-1)/n*c1.*I1(ifwd,n-1);
    fac = -fac;
    I3(ifwd,n+1) = I1(ifwd,n-1)-b1.*I3(ifwd,n)-c1.*I3(ifwd,n-1);
    I5(ifwd,n+1) = I3(ifwd,n-1)-b1.*I5(ifwd,n)-c1.*I5(ifwd,n-1);
  end
  n = N+2;
  I3(ifwd,n+1) = I1(ifwd,n-1)-b1.*I3(ifwd,n)-c1.*I3(ifwd,n-1);
  I5(ifwd,n+1) = I3(ifwd,n-1)-b1.*I5(ifwd,n)-c1.*I5(ifwd,n-1);
end

if ~isempty(ibwd)
  % ...backward recursion
  i1 = zeros(size(ibwd,1),2);
  i3 = zeros(size(ibwd,1),3);
  i5 = zeros(size(ibwd,1),3);
  fac = -1;
  b2c2 = b2./c2;

  % warm-up to get accurate start-values
  for n = 2*(N+3):-1:N+2
    i1 = [(ub2-fac*ua2)./c2/(n-1)+ ...
         (1-2*n)/2/(n-1)*b2c2.*i1(:,1)-n/(n-1)*i1(:,2)./c2 ...
          i1(:,1)];
    fac = -fac;
    i3 = [i1(:,1)./c2-b2c2.*i3(:,1)-i3(:,2)./c2 ...
          i3(:,1:2)];
    i5 = [i3(:,1)./c2-b2c2.*i5(:,1)-i5(:,2)./c2 ...
          i5(:,1:2)];
  end

  % store values from now on
  I1(ibwd,N+1:N+2) = i1(:,1:2);
  I3(ibwd,N+1:N+3) = i3;
  I5(ibwd,N+1:N+3) = i5;
  for n = N+1:-1:3
    I1(ibwd,n-1) = (ub2-fac*ua2)./c2/(n-1)+ ...
        (1-2*n)/(2*(n-1))*b2c2.*I1(ibwd,n)-n/(n-1)*I1(ibwd,n+1)./c2;
    fac = -fac;
    I3(ibwd,n-1) = I1(ibwd,n-1)./c2-b2c2.*I3(ibwd,n)-I3(ibwd,n+1)./c2;
    I5(ibwd,n-1) = I3(ibwd,n-1)./c2-b2c2.*I5(ibwd,n)-I5(ibwd,n+1)./c2;
  end
end

% build Legendre polynomials
L01 = I1(:,1:end-1)*M2L';
L03 = I3(:,1:end-2)*M2L';
L13 = I3(:,2:end-1)*M2L';
L23 = I3(:,3:end)*M2L';
L05 = I5(:,1:end-2)*M2L';
L15 = I5(:,2:end-1)*M2L';
L25 = I5(:,3:end)*M2L';

% output format
iG = cat(3,L01,L03,L05,L13,L15,L23,L25);

% method using forward recursion in Legendre polynomials directly
% (stability issues for c large)
if nargout > 1
  % allocate
  L0 = cell(1,3);

  % initial data
  L0{1} = zeros(len,N+3);
  L0{1}(:,1) = reallog(abs((2+b+2*ub)./(-2+b+2*ua)));

  L0{2} = zeros(len,N+3);
  L0{2}(:,1) = (1+0.5*b)./(d.*ub)-(-1+0.5*b)./(d.*ua);
  L0{2}(izero,1) = -0.5./(1+0.5*b(izero,1)).^2+ ...
      0.5./(-1+0.5*b(izero,1)).^2;

  L0{3} = zeros(len,N+3);
  L0{3}(:,1) = (1+0.5*b)./(3*d.*ub.^3)- ...
      (-1+0.5*b)./(3*d.*ua.^3)+ ...
      2./(3*d).*L0{2}(:,1);
  L0{3}(izero,1) = -0.25./(1+0.5*b(izero,1)).^4+ ...
      0.25./(-1+0.5*b(izero,1)).^4;

  i = 0;
  for beta = 1
    i = i+1;
    for k = 0:N+1
      fac = (k+1)/(2*k+1)*(1+k/(2-beta));
      C = 1/(2-beta)*(ub.^(2-beta)-(-1)^k.*ua.^(2-beta));
      ib = k-2:-2:0;
      bterm = L0{i}(:,ib+1)*(2*ib+1)';
      ic = k-1:-2:0;
      cterm = L0{i}(:,ic+1)*(2*ic+1)';

      L0{i}(:,k+2) = ...
          1/fac*(C- ...
                 b*(0.5+k/(2-beta)).*L0{i}(:,k+1)- ...
                 (k/(2*k+1)-1/(2-beta)*k*(k+1)/(2*k+1))*L0{i}(:,max(1,k))- ...
                 b/(2-beta).*bterm- ...
                 (c+1)/(2-beta).*cterm);
    end
  end
  % recursion above is singular for beta = {3,5} (fac will be zero)
  for beta = [3 5]
    i = i+1;
    for k = 0:N+1
      fac = (2*k+1)/(k+1);
      C = 1/(2-beta)*(ub.^(2-beta)-(-1)^k.*ua.^(2-beta));
      il = k-1:-2:0;
      Lterm = L0{i-1}(:,il+1)*(2*il+1)';

      L0{i}(:,k+2) = fac*(C- ...
                          b/2.*L0{i}(:,k+1)- ...
                          k/(2*k+1)*L0{i}(:,max(1,k))- ...
                          1/(2-beta)*Lterm);
    end
  end

  % construct alpha = {1,2}-cases from the case alpha = 0 
  k = 0:N+1;
  k1 = (k+1)./(2*k+1);
  k2 = k./(2*k+1);
  L01 = L0{1}(:,1:N+1);
  L03 = L0{2}(:,1:N+1);
  L13 = tprod(k1,L0{2}(:,2:end),[3 2],[1 2])+ ...
        tprod(k2,L0{2}(:,[1 1:end-2]),[3 2],[1 2]);
  L23 = tprod(k1(1:end-1),L13(:,2:end),[3 2],[1 2])+ ...
        tprod(k2(1:end-1),L13(:,[1 1:end-2]),[3 2],[1 2]);
  L13 = L13(:,1:N+1);
  L05 = L0{3}(:,1:N+1);
  L15 = tprod(k1,L0{3}(:,2:end),[3 2],[1 2])+ ...
        tprod(k2,L0{3}(:,[1 1:end-2]),[3 2],[1 2]);
  L25 = tprod(k1(1:end-1),L15(:,2:end),[3 2],[1 2])+ ...
        tprod(k2(1:end-1),L15(:,[1 1:end-2]),[3 2],[1 2]);
  L15 = L15(:,1:N+1);

  % build output
  jG = cat(3,L01,L03,L05,L13,L15,L23,L25);
end

%% series expansion for I1, converges quickly for c large
%iunstable = find(realsqrt(c).^(N+3) > 1e6);
%if ~isempty(iunstable)
%  bb = b(iunstable);
%  cc = c(iunstable);
%
%  % number of terms in series, err ~ c^(-Nterm)
%  Nterm = round(-reallog(1e-8)/reallog(min(cc)));
%
%  % quick and dirty way:
%  kk = 0:N;
%  nn = tsum(kk,2*(0:Nterm-1),[3 2],[3 1]);
%  kk = frepmat(kk,Nterm);
%  C = sqrt(pi)*exp(-log(2)*nn+gammaln(nn+1)- ...
%                   gammaln(1+(nn-kk)/2)-gammaln(1.5+(nn+kk)/2));
%
%  % evaluate sum
%  P = legeval(-0.5*bb./realsqrt(cc),max(nn(:)));
%  I1ser = tprod(reshape(P(nn+1,:),Nterm,N+1,[]).* ...
%                exp(tprod(-0.5*reallog(cc(:)),nn+1,3,1:2)), ...
%                C,[-1 1 2],[-1 1]);
%end
