function ipG = greeninterp(x,y,z,G)
%GREENINTERP Interpolation of Green's function.
%   IPG = GREENINTERP(X,Y,Z,G) interpolates in the Green's function
%   table G at the coordinates X, Y, Z. The result is returned in a
%   3-D array with dimensions size(X,1)-by-size(X,2)-by-6.
%
%   For an explanation of the input G, see GREENTABLE.
%
%   Performance of this function is important. No error-checking is
%   performed.
%
%   See also GREENTABLE.

% S. Engblom 2009-06-17 (Revision, Chebyshev format)
% S. Engblom 2009-05-11 (Revision)
% A-K. Tornberg (original codes 'ip_Gp' and 'trilin_ip_6_vec')

% since table is all positive
sx = sign(x);
sy = sign(y);
sz = sign(z);
x = abs(x);
y = abs(y);
z = abs(z);

% linear/cubic interpolation
if ~isscalar(G.ntable)
  % find table entries
  x = x/G.ntable(2);
  y = y/G.ntable(2);
  z = z/G.ntable(2);
  ix = floor(x);
  iy = floor(y);
  iz = floor(z);

  % local coordinates in [0,1]
  x = x-ix;
  y = y-iy;
  z = z-iz;

  % interpolate
  if G.ntable(1) == 2
    ipG = l_interp3lin(x,y,z,ix,iy,iz,G.G,G.sizG);
  else
    ipG = l_interp3cub(x,y,z,ix,iy,iz,G.G,G.sizG);
  end
% Chebyshev interpolation
else
  % local coordinates in [-1,1]
  x = 2*(x/G.Lmax(1))-1;
  y = 2*(y/G.Lmax(2))-1;
  z = 2*(z/G.Lmax(3))-1;

  % evaluate
  ipG = l_interpcheb(x,y,z,G.G);
end

% fix sign and reshape
ipG = reshape(ipG,[size(x) 6]);
ipG(:,:,4) = sx.*sy.*ipG(:,:,4);
ipG(:,:,5) = sy.*sz.*ipG(:,:,5);
ipG(:,:,6) = sx.*sz.*ipG(:,:,6);

%-------------------------------------------------------------------------
function ipG = l_interp3lin(x,y,z,ix,iy,iz,G,sizG)
%L_INTERP3LIN Trilinear interpolation.
%   IPG = L_INTERP3LIN(X,Y,Z,IX,IY,IZ,G,sizG) interpolates linearly in
%   the table G at the local coordinates X, Y, Z (in [0,1]) found in G
%   at the indices IX, IY and IZ, respectively. The result is returned
%   in a matrix of size numel(X)-by-size(G,2).

% conjugate coordinates
x = x(:)';
y = y(:)';
z = z(:)';
x2 = 1-x;
y2 = 1-y;
z2 = 1-z; 

% build interpolation coefficients
xyz = x.*y.*z;
xyz2 = x2.*y2.*z2;
IM = [xyz2; ...
      y2.*z2-xyz2; x2.*z2-xyz2; x2.*y2-xyz2; ...
      y.*z-xyz; x.*z-xyz; x.*y-xyz; ...
      xyz];

% indices to interpolate from
strG = [1 cumprod(sizG(1:end-1))]';
sg = sum(strG);
shift = [0; strG; sg-strG; sg];
IX = tsum(shift,1+[ix(:) iy(:) iz(:)]*strG,1,2);

% evaluate
ipG = tprod(IM,reshape(G(IX,:),8,size(IX,2),[]),[-1 1],[-1 1 2]);
% (this line of code is critical to execution speed)

%-------------------------------------------------------------------------
function ipG = l_interp3cub(x,y,z,ix,iy,iz,G,sizG)
%L_INTERP3cub 4th order (cubic) interpolation.
%   IPG = L_INTERP3cub(X,Y,Z,IX,IY,IZ,G,sizG) works like
%   L_INTERP3LIN but implements a 4th order interpolation formula.

% formula built using points as in [0  1 * 2  3]
strG = [1 cumprod(sizG(1:end-1))]';
shift = (0:3)'*strG';
shift = tsum(tsum(shift(:,1),shift(:,2),1,2),shift(:,3),[1 2],3);
IX = tsum(shift(:),1+[ix(:) iy(:) iz(:)]*strG,1,2); % 64-by-numel(x)

% interpolation coefficients
p = x(:)';
p_1 = p+1;
p1 = p-1;
p2 = p-2;
% (note: here the coordinates are [-1 0 1 2])
IM1 = [-1/6*p.*p1.*p2; ...
       1/2*p1.*p2.*p_1; ...
       -1/2*p.*p_1.*p2; ...
       1/6*p.*p1.*p_1];
p = y(:)';
p_1 = p+1;
p1 = p-1;
p2 = p-2;
IM2 = [-1/6*p.*p1.*p2; ...
       1/2*p1.*p2.*p_1; ...
       -1/2*p.*p_1.*p2; ...
       1/6*p.*p1.*p_1];
p = z(:)';
p_1 = p+1;
p1 = p-1;
p2 = p-2;
IM3 = [-1/6*p.*p1.*p2; ...
       1/2*p1.*p2.*p_1; ...
       -1/2*p.*p_1.*p2; ...
       1/6*p.*p1.*p_1];

IM = tprod(tprod(IM1,IM2,[1 3],[2 3]),IM3,[1 2 4],[3 4]);
ipG = tprod(reshape(IM,64,[]),reshape(G(IX,:),64,[],6),[-1 1],[-1 1 2]);
% (major computation)

% $$$ % iterated evaluation
% $$$ p = x(:)';
% $$$ p_1 = p+1; % (note: here the coordinates are [-1 0 1 2])
% $$$ p1 = p-1;
% $$$ p2 = p-2;
% $$$ IM = [-1/6*p.*p1.*p2; ...
% $$$       1/2*p1.*p2.*p_1; ...
% $$$       -1/2*p.*p_1.*p2; ...
% $$$       1/6*p.*p1.*p_1];
% $$$ ipG = tprod(IM,reshape(G(IX,:),4,16,[],6),[-1 2],[-1 1:3]);
% $$$ % (the major computation)
% $$$ 
% $$$ p = y(:)';
% $$$ p_1 = p+1;
% $$$ p1 = p-1;
% $$$ p2 = p-2;
% $$$ IM = [-1/6*p.*p1.*p2; ...
% $$$       1/2*p1.*p2.*p_1; ...
% $$$       -1/2*p.*p_1.*p2; ...
% $$$       1/6*p.*p1.*p_1];
% $$$ ipG = tprod(IM,reshape(ipG,4,4,[],6),[-1 2],[-1 1:3]);
% $$$ 
% $$$ p = z(:)';
% $$$ p_1 = p+1;
% $$$ p1 = p-1;
% $$$ p2 = p-2;
% $$$ IM = [-1/6*p.*p1.*p2; ...
% $$$       1/2*p1.*p2.*p_1; ...
% $$$       -1/2*p.*p_1.*p2; ...
% $$$       1/6*p.*p1.*p_1];
% $$$ ipG = tprod(IM,reshape(ipG,4,[],6),[-1 1],[-1 1 2]);

%-------------------------------------------------------------------------
function ipG = l_interpcheb(x,y,z,G)
%L_INTERPCHEB Chebyshev spectral interpolation.
%   IPG = L_INTERPCHEB(X,Y,Z,G) evaluates the Chebyshev approximation
%   stored as coefficients in G at the local coordinates X, Y, Z (in
%   [-1,1]). The result is returned in a matrix of size
%   numel(X)-by-size(G,4).
%
%   It is required that all(tsize(G,1:3) > 1) so that the
%   interpolation is at least linear.

% x-coordinates
A = [zeros(1,numel(x)); x(:)'; frepmat(2*x(:)',size(G,1)-2)];
B = [1; 0; frepmat(-1,max(tsize(G,1:3))-2)];
ipG = clenshaw(A,B(1:size(G,1)),reshape(G,size(G,1),[]));
% tsize(G,2:4)-by-numel(x)

% y-coordinates
A = [zeros(1,numel(y)); y(:)'; frepmat(2*y(:)',size(G,2)-2)];
ipG = clenshaw(A,B(1:size(G,2)),reshape(ipG,size(G,2),[],numel(x)));
% tsize(G,3:4)-by-numel(x)

% z-coordinates
A = [zeros(1,numel(z)); z(:)'; frepmat(2*z(:)',size(G,3)-2)];
ipG = clenshaw(A,B(1:size(G,3)),reshape(ipG,size(G,3),[],numel(x)));
% size(G,4)-by-numel(x)

% *** prefer to use a different normalization?

% required format
ipG = ipG'; % *** can get rid of this

%-------------------------------------------------------------------------
