function G = greenstokes(X,Y,Z,Ldim,varargin)
%GREENSTOKES Periodic Green's function remainder (Stokeslet).
%   G = GREENSTOKES(X,Y,Z,Ldim,...) computes the periodic Green's
%   function for Stokes flow, with the Stokeslet for the center
%   location subtracted off. X, Y, Z are the coordinates and Ldim is
%   the size of the periodic box.
%
%   If Ldim = [Lx Ly Lz] (or [Lx Ly]), then periodicity in the
%   corresponding 3 (or 2) dimensions is understood.
%
%   The output G is a 4-D array of size
%   numel(X)-by-numel(Y)-by-numel(Z)-by-6 and this format is to be
%   understood as follows:
%     G(:,:,:,1) = G_XX
%     G(:,:,:,2) = G_YY
%     G(:,:,:,3) = G_ZZ
%     G(:,:,:,4) = G_XY
%     G(:,:,:,5) = G_YZ
%     G(:,:,:,6) = G_XZ
%   This enumeration can be thought of as a banded format of the upper
%   triangular part of the full tensor G.
%
%   Property     Value/{Default}         Description
%   -----------------------------------------------------------------------
%   ksi          Scalar                  Ewald-summation parameter.
%
%   rlayers,     Scalar {8}              Number of layers in 
%   klayers                              real- and reciprocal 
%                                        space, respectively.
%
%   blocksize    {1e6}                   Vectorization blocksize.
%
%   report       Boolean {false}         Usage of reporter.
%
%   The default value of the property ksi is
%   sqrt(pi)/(tau*max(Ldim))^(1/4), where tau = prod(Ldim) is the
%   volume of the periodic box.
%
%   Performance of this function is important. Only very limited
%   error-checking is performed.
%
%   See also GREENTABLE.

% Reference:
%   C. Pozrikidis: "Computation of periodic Green's functions of
%   Stokes flow", J. Eng. Math. 30:79--96 (1996).

% S. Engblom 2009-05-11 (Revision)
% A-K. Tornberg (original code 'Gp_erf')

% options
tau = prod(Ldim);
opts.ksi = realsqrt(pi/realsqrt(tau*max(Ldim)));
opts.rlayers = 8;
opts.klayers = 8;
opts.blocksize = 1e6;
opts.report = 0;
opts = parseopts(opts,varargin);
ksi = opts.ksi;

% allocate output
G = zeros(numel(X)*numel(Y)*numel(Z),6);

% reporter, if any
if opts.report
  report([1 2*size(G,1)],'Evaluating Green''s function remainder...','init');
else
  report([],[],'none');
end

% |r| < tolzero are understood as 0 and the limit is calculated
tolzero = eps(1e6);

if numel(Ldim) == 3 % S3D3P: implements 3.1 in [1]
  % (1) sum in real space

  % determine suitable size of blocks
  block = max(1,round(opts.blocksize/(2*opts.rlayers+1)^3));
  % (vectorized over layers)

  % coordinate grid
  layers = -opts.rlayers:opts.rlayers;
  XP = tsum(X(:),layers*Ldim(1),1,[3 2]);
  YP = tsum(Y(:),layers*Ldim(2),1,[3 2]);
  ZP = tsum(Z(:),layers*Ldim(3),1,[3 2]);

  % norms
  XP2 = XP.*XP;
  YP2 = YP.*YP;
  ZP2 = ZP.*ZP;

  % vectorize over chunks of coordinates
  for begin = 1:block:size(G,1)
    iG = begin:min(begin+block-1,size(G,1));
    [ix,iy,iz] = ind2sub([numel(X) numel(Y) numel(Z)],iG(:));
    
    % norms
    Rmod3 = tsum(tsum(XP2(ix,:),YP2(iy,:),[1 2],[1 3]), ...
                 ZP2(iz,:),[1 2 3],[1 4]);
    Rmod = realsqrt(Rmod3);
    x = Rmod*ksi;
    x2 = x.*x;
    
    % careful at zero:
    ixzero = find(Rmod(:,opts.rlayers+1, ...
                       opts.rlayers+1, ...
                       opts.rlayers+1) < tolzero);
    Rmod = max(Rmod,tolzero);
    Rmod3 = max(Rmod3,tolzero).*Rmod;
    
    % core functions
    T0 = erfc(x);
    T1 = 4/realsqrt(pi)*x.*exp(-x2);
    T2 = (x2-0.5).*T1;
    
    % correction at origo:
    T0(:,opts.rlayers+1,opts.rlayers+1,opts.rlayers+1) = ...
        T0(:,opts.rlayers+1,opts.rlayers+1,opts.rlayers+1)-1;
    
    % partial sums
    C = tsum((T0+T2-T1)./Rmod,2:4);
    D = (T0-T2)./Rmod3;
    D4 = sum(D,4);
    D3 = reshape(sum(D,3),size(D4));
    D2 = reshape(sum(D,2),size(D4));
    
    % final sums
    G(iG,1) = C+tprod(sum(D4,3),XP2(ix,:),[1 -1],[1 -1]);
    G(iG,2) = C+tprod(sum(D2,3),YP2(iy,:),[1 -1],[1 -1]);
    G(iG,3) = C+tprod(reshape(sum(D3,2),tsize(D3,[1 3])),ZP2(iz,:), ...
                      [1 -1],[1 -1]);
    G(iG,4) = tprod(tprod(D4,XP(ix,:),[1 -1 2],[1 -1]), ...
                    YP(iy,:),[1 -1],[1 -1]);
    G(iG,5) = tprod(tprod(D2,YP(iy,:),[1 -1 2],[1 -1]), ...
                    ZP(iz,:),[1 -1],[1 -1]);
    G(iG,6) = tprod(tprod(D3,XP(ix,:),[1 -1 2],[1 -1]), ...
                    ZP(iz,:),[1 -1],[1 -1]);
    
    % add limit term C --> -8*ksi/sqrt(pi) as Rmod --> 0 (the limit of the
    % D-term is zero)
    G(iG(ixzero),1:3) = G(iG(ixzero),1:3)-8*ksi/realsqrt(pi);
    report(begin,0,'');
  end
  
  % (2) sum in reciprocal space
  block = max(1,round(opts.blocksize/(2*opts.klayers+1)^3));
  
  % coordinate grid
  layers = -opts.klayers:opts.klayers;
  XP = layers*2*pi/Ldim(1);
  YP = layers*2*pi/Ldim(2);
  ZP = layers*2*pi/Ldim(3);

  % norms
  XP2 = XP.*XP;
  YP2 = YP.*YP;
  ZP2 = ZP.*ZP;
  Kmod2 = tsum(tsum(XP2,YP2,[3 1],[3 2]),ZP2,[1 2],[4 3]);
  Kmod = realsqrt(Kmod2);
  % correction at origo:
  Kmod(opts.klayers+1,opts.klayers+1,opts.klayers+1) = 1;
  
  omega = Kmod/ksi;
  omega2 = omega.*omega;
  T0 = 8*pi/tau./realpow(Kmod,4).* ...
       (((0.125*omega2+0.25)).*omega2+1).*exp(-0.25*omega2);
  
  for begin = 1:block:size(G,1)
    iG = begin:min(begin+block-1,size(G,1));
    [ix,iy,iz] = ind2sub([numel(X) numel(Y) numel(Z)],iG(:));
    
    T1 = cos(tsum(tsum(X(ix)*XP,Y(iy)*YP,[1 2],[1 3]), ...
                  Z(iz)*ZP,[1 2 3],[1 4]));
    C = tprod(T0,T1,[2 3 4],1:4);
    % correction at origo:
    C(:,opts.klayers+1,opts.klayers+1,opts.klayers+1) = 0;
    C4 = reshape(sum(C,4),size(C,1),[]);
    C3 = reshape(sum(C,3),size(C,1),[]);
    C2 = reshape(sum(C,2),size(C,1),[]);
    C = reshape(C,size(C,1),[]);
    
    % final sums
    G(iG,1) = G(iG,1)+C*reshape(tsum(Kmod2,-XP2,1:3,[4 1]),[],1);
    G(iG,2) = G(iG,2)+C*reshape(tsum(Kmod2,-YP2,1:3,[4 2]),[],1);
    G(iG,3) = G(iG,3)+C*reshape(tsum(Kmod2,-ZP2,1:3,[4 3]),[],1);
    G(iG,4) = G(iG,4)-C4*reshape(XP'*YP,[],1);
    G(iG,5) = G(iG,5)-C2*reshape(YP'*ZP,[],1);
    G(iG,6) = G(iG,6)-C3*reshape(XP'*ZP,[],1);
    report(size(G,1)+begin,[],'');
  end
elseif numel(Ldim) == 2 % S3D2P: implements 3.2 in [1]
  error('Not yet implemented.'); % ***

  % (1) sum in real space; this code is similar to S3D3P above but with
  % only one Z-layer (namely 0)

  % determine suitable size of blocks
  block = max(1,round(opts.blocksize/(2*opts.rlayers+1)^2));
  % (vectorized over layers)

  % coordinate grid
  layers = -opts.rlayers:opts.rlayers;
  XP = tsum(X(:),layers*Ldim(1),1,[3 2]);
  YP = tsum(Y(:),layers*Ldim(2),1,[3 2]);
  ZP = Z(:);

  % norms
  XP2 = XP.*XP;
  YP2 = YP.*YP;
  ZP2 = ZP.*ZP;

  % vectorize over chunks of coordinates
  for begin = 1:block:size(G,1)
    iG = begin:min(begin+block-1,size(G,1));
    [ix,iy,iz] = ind2sub([numel(X) numel(Y) numel(Z)],iG(:));
    
    % norms
    Rmod3 = tsum(tsum(XP2(ix,:),YP2(iy,:),[1 2],[1 3]), ...
                 ZP2(iz),[1 2 3],1);
    Rmod = realsqrt(Rmod3);
    x = Rmod*ksi;
    x2 = x.*x;
    
    % careful at zero:
    ixzero = find(Rmod(:,opts.rlayers+1, ...
                       opts.rlayers+1) < tolzero);
    Rmod = max(Rmod,tolzero);
    Rmod3 = max(Rmod3,tolzero).*Rmod;

    % core functions
    T0 = erfc(x);
    T1 = 4/realsqrt(pi)*x.*exp(-x2);
    T2 = (x2-0.5).*T1;
    
    % correction at origo:
    T0(:,opts.rlayers+1,opts.rlayers+1) = ...
        T0(:,opts.rlayers+1,opts.rlayers+1)-1;

    % partial sums
    C = tsum((T0+T2-T1)./Rmod,2:4);
    D = (T0-T2)./Rmod3;
    D3 = sum(D,3);
    D2 = reshape(sum(D,2),size(D3));

    % final sums
    G(iG,1) = C+tprod(D3,XP2(ix,:),[1 -1],[1 -1]);
    G(iG,2) = C+tprod(D2,YP2(iy,:),[1 -1],[1 -1]);
    G(iG,3) = C+sum(D3,2).*ZP2(iz);
    G(iG,4) = tprod(tprod(D,XP(ix,:),[1 -1 2],[1 -1]), ...
                    YP(iy,:),[1 -1],[1 -1]);
    G(iG,5) = tprod(D2,YP(iy,:),[1 -1],[1 -1]).*ZP(iz);
    G(iG,6) = tprod(D3,XP(ix,:),[1 -1],[1 -1]).*ZP(iz);

    % add limit term C --> -8*ksi/sqrt(pi) as Rmod --> 0 (the limit of the
    % D-term is zero)
    G(iG(ixzero),1:3) = G(iG(ixzero),1:3)-8*ksi/realsqrt(pi);
    report(begin,0,'');
  end

  % (2) sum in reciprocal space
  block = max(1,round(opts.blocksize/(2*opts.klayers+1)^2));

  % coordinate grid (2-D sum)
  layers = -opts.klayers:opts.klayers;
  XP = layers*2*pi/Ldim(1);
  YP = layers*2*pi/Ldim(2);

  % norms
  XP2 = XP.*XP;
  YP2 = YP.*YP;
  Lmod2 = tsum(XP2,YP2,[3 1],[3 2]);
  Lmod = realsqrt(Lmod2);
  % correction at origo:
  Lmod(opts.klayers+1,opts.klayers+1) = 1;

  zeta = Lmod/ksi;
  zeta2 = zeta.*zeta;
  T0 = 8/tau/ksi^3*exp(-0.25*zeta2);

  for begin = 1:block:size(G,1)
    iG = begin:min(begin+block-1,size(G,1));
    [ix,iy,iz] = ind2sub([numel(X) numel(Y) numel(Z)],iG(:));

    delta = abs(ksi*Z(iz));
    delta2 = delta.*delta;
    rho = reshape(delta*zeta(:)',[],2*opts.klayers+1,2*opts.klayers+1);

    % integrals *** can be optimized!
    I0 = realsqrt(pi)*exp(-delta2);

    i1 = exp(rho).*erf(tsum(0.5*zeta,delta,[2 3],1));
    i2 = exp(-rho).*erf(tsum(0.5*zeta,-delta,[2 3],1));

    I1 = tprod(-pi/4./zeta.*exp(0.25*zeta2),i1+i2-2*cosh(rho),[2 3],1:3);

    I2 = tprod(1/2./zeta2-0.25,I1,[2 3],1:3)+ ...
         tprod(pi/8./zeta2.*exp(0.25*zeta2), ...
               tprod(delta,i1-i2-2*sinh(rho),1,1:3)+ ...
               tprod(2/realsqrt(pi)*exp(0.5*zeta), ...
                     cosh(tsum(rho,delta,1:3,1)),[2 3],1:3), ...
               [2 3],1:3);

    T1 = tsum(I0,I1+I2,1,1:3).*cos(tsum(X(ix)*XP,Y(iy)*YP,[1 2],[1 3]));

    % *** d/dz and d^2/dz^2

    C = tprod(T0,T1,[2 3],1:3);
    % correction at origo:
    C(:,opts.klayers+1,opts.klayers+1) = 0;
    C3 = reshape(sum(C,3),size(C,1),[]);
    C2 = reshape(sum(C,2),size(C,1),[]);
    % correction at origo:
    % *** add the zero term

    % final sums
    G(iG,1) = G(iG,1)+tprod(C,YP2,[1 2 -1],[3 -1]); % ***
    G(iG,2) = G(iG,2)+tprod(C,XP2,[1 -1 2],[3 -1]); % ***
    G(iG,3) = G(iG,3)+reshape(C,size(C,1),[])*Lmod2(:);
    G(iG,4) = G(iG,4)+reshape(C,size(C,1),[])*reshape(XP'*YP,[],1);
    G(iG,5) = G(iG,5); % *** 
    G(iG,6) = G(iG,6); % ***
    report(size(G,1)+begin,[],'');
  end
else
  error('Only 3- and 2-periodic Stokeslet supported.');
end

% reshape to 4-D
G = reshape(G,numel(X),numel(Y),numel(Z),6);

% close reporter
report([],[],'done');
