function G = greentable(Ldim,Lmax,ntable,varargin)
%GREENTABLE Table of Green's function remainder (Stokeslet).
%   G = GREENTABLE(Ldim,Lmax,ntable) tabulates the periodic Green's
%   function for Stokes flow, with the Stokeslet for the center
%   location subtracted off. Ldim = [Lx Ly Lz] is the dimension of the
%   periodic box and Lmax = [xmax ymax zmax] the ranges for tabulation
%   (strarting at 0).
%
%   Two syntaxes are supported: if ntable is a length 2 vector,
%   ntable(2) is used as the discretization step and Lagrange
%   interpolation of order ntable(1) = 2 (linear) or ntable(1) = 4
%   (cubic) will be used by GREENINTERP. Otherwise, ntable(1)-1 >= 1
%   is the highest order Chebyshev polynomial used for representing
%   the remainder.
%
%   On return, G is the Green's function table (for format of this
%   structure, see FIBRSETUP).
%
%   The syntax GREENTABLE(Ldim,Lmax,ntable,...) allows for additional
%   options to be passed directly to GREENSTOKES.
%
%   See also GREENSTOKES, GREENINTERP, FIBRSETUP.

% S. Engblom 2009-06-17 (Revision, Chebyshev format)
% S. Engblom 2009-05-11 (Revision)
% A-K. Tornberg (original code 'tabulate_Gp_posloc')

% 32-bit checksum
G.check = fsetop('check',[Ldim(:); Lmax(:); ntable(:)]);
G.ntable = ntable;
G.Lmax = Lmax;

% constant step-size table
if ~isscalar(ntable)
  % 3-D mesh
  h = ntable(2);
  N = ntable(1);
  G.sizG = ceil(Lmax/h)+1;
  Lmax = h*(G.sizG-1); % effective box might be a little bit larger
  if N == 2
    % linear case
    X = 0:h:Lmax(1);
    Y = 0:h:Lmax(2);
    Z = 0:h:Lmax(3);
  elseif N == 4
    % cubic case
    X = -h:h:Lmax(1)+h;
    Y = -h:h:Lmax(2)+h;
    Z = -h:h:Lmax(3)+h;
    G.sizG = G.sizG+2;
  else
    error('Unknown interpolation.');
  end

  % check
  if any([X(end) Y(end) Z(end)] >= Ldim)
    warning('Tabulating at singularity.');
  end

  % single build
  G.G = greenstokes(X(:),Y(:),Z(:),Ldim,varargin{:});

  % G as a matrix is more convenient
  G.G = reshape(G.G,[],6);
else % Chebyshev expansion
  % quadrature
  N = ntable(1);
  Nq = N+1;
  xx = -cos((2*(1:Nq)-1)*pi/2/Nq);
  ww = pi/Nq;

  % evaluate Chebyshev polynomials
  j = [0; 1; 2*ones(N-2,1)];
  A = j*xx;
  B = [1; 0; -ones(N-2,1)];
  T = clenshaw(A,B);

  % quadrature nodes for each dimension
  X = 0.5*(xx+1)*Lmax(1);
  Y = 0.5*(xx+1)*Lmax(2);
  Z = 0.5*(xx+1)*Lmax(3);

  % evaluate
  Gf = greenstokes(X(:),Y(:),Z(:),Ldim,varargin{:});
  
  % inner product against polynomials
  T = ww*T;
  Gf = tprod(T,Gf,[1 -1],[-1 2:4]);
  Gf = tprod(T,Gf,[2 -1],[1 -1 3:4]);
  Gf = tprod(T,Gf,[3 -1],[1:2 -1 4]);

  % Chebyshev coefficients
  G.G = Gf/(pi/2)^3;
  G.G(1,:,:,:) = 0.5*G.G(1,:,:,:);
  G.G(:,1,:,:) = 0.5*G.G(:,1,:,:);
  G.G(:,:,1,:) = 0.5*G.G(:,:,1,:);
end
