function varargout = randfibr(varargin)
%RANDFIBR Random initialization of fibers.
%   [X,T] = RANDFIBR(M,Ldim,rlim) initializes M fibers randomly in a
%   box of dimensions Ldim = [Lx Ly Lz] with the requirement not to
%   include any fiber that is within a distance smaller than rlim from
%   any other fiber. The output includes position and (normalized)
%   tangent matrices, respectively, of sizes 3-by-M.
%
%   INIT = RANDFIBR(FIBR,Ldim) is an alternative syntax where the
%   inputs are obtained from the FIBR-structure FIBR. Here all fibers
%   are contained in a box [-Ldim/2,Ldim/2] (by shifting).
%
%   INIT = RANDFIBR(FIBR) uses Ldim = FIBR.Ldim.
%
%   RANDFIBR changes the state of the random number generator.
%
%   See also FIBRSETUP.

% S. Engblom 2010-01-27 (2-argument syntax added)
% S. Engblom 2009-06-08 (Minor revision)
% S. Engblom 2009-04-30 (Major revision)
% A-K. Tornberg 2009-02-25 (original code 'initialize_fibers_random')

% parse input
if nargin < 3
  % quick check to avoid strange error messages
  fibr = varargin{1};
  if fibr.check ~= fibrcheck(fibr)
    error('Struct has changed, call FIBRSETUP before solving.');
  end
  if nargin == 1
    M = fibr.Mfibers;
    Ldim = fibr.Ldim;
    rlim = fibr.rlim;
  else
    M = fibr.Mfibers;
    Ldim = varargin{2};
    rlim = fibr.rlim;
  end
else
  M = varargin{1};
  Ldim = varargin{2};
  rlim = varargin{3};
end
% being very pedagogical here...
if numel(Ldim) ~= 3 || ~isscalar(rlim)
  error(['Box must be defined by 3 coordinates and minimal fiber ' ...
         'distance should be a scalar.']);
end

% first fiber
X = Ldim(:).*rand(3,1);
T = rand(3,1);
T = T/norm(T);

% Algorithm: start adding random fibers to the output array. For each
% new fiber to add, check that it is at least rlim away from all the
% fibers already added.
while  size(X,2) < M
  % new tentative random fiber
  xx = Ldim(:).*rand(3,1);
  tt = rand(3,1);
  tt = tt/norm(tt);

  % add the fiber if possible
  D = distances([xx; tt],[X; T],rlim);
  if nnz(D) == 0
    X = [X xx];
    T = [T tt];
  end
end

% parse output
if nargout == 1
  % prefer symmetric coordinates:
  X = tsum(X,-Ldim/2,[1 2],[3 1]);
  varargout{1} = [X; T];
else
  varargout{1} = X;
  varargout{2} = T;
end
